package compiler.ast;

/**
 * Resolve unresolved symbols to their actual symbol values.
 */
public class ResolveSymbolVisitor extends AbstractVoidVisitor {
  private SymbolTable table;

  private String getId(Symbol s) {
    return ((UnresolvedSymbol) s).getId();
  }

  public ResolveSymbolVisitor(SymbolTable table) {
    this.table = table;
  }

  public void visit(Assign n) {
    Symbol s = table.getVariableSymbol(getId(n.getName()));

    n.setName(s);
    n.getValue().accept(this);
  }

  public void visit(Call n) {
    Symbol s = table.getMethodSymbol(getId(n.getName()));
    
    n.setName(s);

    for (Node actual : n.getActuals())
      actual.accept(this);
  }

  public void visit(FieldDecl n) {
    table.declareField(n.getVarType(), getId(n.getName()));
  }

  public void visit(MethodDecl n) {
    Symbol s = table.getMethodSymbol(getId(n.getName()));

    n.setName(s);
    table.pushMethodContext(s);

    for (VarDecl f : n.getFormals())
      table.declareFormal(f.getVarType(), getId(f.getName()));

    n.getBody().accept(this);

    table.popContext();
  }

  public void visit(Var n) {
    Symbol s = table.getVariableSymbol(getId(n.getName()));
    n.setName(s);
  }

  public void visit(VarDecl n) {
    table.declareLocal(n.getVarType(), getId(n.getName()));
  }
}
