package compiler.ast;

import java.util.Map;
import java.util.HashMap;

/**
 * Type factory.
 */
public class Type {
  private static Type floatType = new Type();
  private static Type voidType = new Type();
  private static Type intType = new Type();
  private static Type booleanType = new Type();
  private static Map<TypePair,Type> functionTable = new HashMap<TypePair,Type>();
  private static Map<TypePair,Type> productTable = new HashMap<TypePair,Type>();


  private static class TypePair {
    private Type t1, t2;
    public TypePair(Type t1, Type t2) { this.t1 = t1; this.t2 = t2; }
    public boolean equals(Object other) {
      if (other instanceof TypePair) {
        TypePair o = (TypePair) other;
        return t1.equalsType(o.t1) && t2.equalsType(o.t2);
      }
      return false;
    }
    public int hashCode() { return t1.hashCode() + t2.hashCode(); }
  }

  private static class ProductType extends Type {
    private Type t1, t2;
    public ProductType(Type t1, Type t2) { this.t1 = t1; this.t2 = t2; }
  }

  private static class FunctionType extends Type {
    private Type from, to;
    public FunctionType(Type from, Type to) { this.from = from; this.to = to; }
    public Type fromType() { return from; }
    public Type toType() { return to; }
  }

  public boolean equalsType(Type t) { return this == t; }
  public Type fromType() { throw new Error("Not function type"); }
  public Type toType() { throw new Error("Not function type"); }

  public static Type productType(Type t1, Type t2) {
    TypePair key = new TypePair(t1, t2);

    if (productTable.containsKey(key))
      return productTable.get(key);

    Type t = new ProductType(t1, t2);
    productTable.put(key, t);
    return t;
  }

  public static Type functionType(Type from, Type to) {
    TypePair key = new TypePair(from, to);
    
    if (functionTable.containsKey(key))
      return functionTable.get(key);

    Type t = new FunctionType(from, to);
    functionTable.put(key, t);
    return t;
  }

  public static Type floatType() { return floatType; }
  public static Type voidType() { return voidType; }
  public static Type intType() { return intType; }
  public static Type booleanType() { return booleanType; }
}
